package net.wanji.utc.hisense.netty;

import io.netty.bootstrap.Bootstrap;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioSocketChannel;
import lombok.extern.slf4j.Slf4j;
import net.wanji.utc.hisense.cache.netty.NettyMessageCache;
import net.wanji.utc.hisense.common.enums.xml.MessageTypeEnum;
import net.wanji.utc.hisense.netty.codec.MessageDecoder;
import net.wanji.utc.hisense.netty.codec.MessageEnCoder;
import net.wanji.utc.hisense.netty.handler.TcpClientHandler;
import net.wanji.utc.hisense.pojo.xml.pojo.messagecontent.HeartBeatRequest;
import net.wanji.utc.hisense.util.XMLUtils;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.Objects;

/**
 * @author duanruiming
 * @date 2024/01/08 17:33
 */
@Slf4j
@Component
public class TcpClient {
    private static ChannelFuture tcpChannelFuture = null;
    @Resource(name = "commonThreadPoolExecutor")
    private ThreadPoolTaskExecutor commonThreadPoolExecutor;

    public static void connection(String host, int port) throws Exception {
        EventLoopGroup group = new NioEventLoopGroup(); // 创建线程组
        try {
            Bootstrap b = new Bootstrap(); // 创建客户端启动对象
            b.group(group) // 设置线程组
                    .channel(NioSocketChannel.class) // 设置通道实现类
                    .handler(new ChannelInitializer<Channel>() { // 设置通道初始化对象
                        @Override
                        protected void initChannel(Channel ch) throws Exception {
                            ch.pipeline().addLast(new MessageDecoder());
                            ch.pipeline().addLast(new MessageEnCoder());
                            ch.pipeline().addLast(new TcpClientHandler()); // 添加自定义的处理器
                        }
                    });
            // 启动客户端，连接到服务器端，并同步，生成ChannelFuture对象
            ChannelFuture f = b.connect(host, port).sync();
            //channelFutureMap.put(f.channel().id(), f);
            tcpChannelFuture = f;
            // 对关闭通道进行监听
            //f.channel().closeFuture().sync();
            Runtime.getRuntime().addShutdownHook(new Thread(() -> {
                group.shutdownGracefully();
                f.channel().closeFuture().syncUninterruptibly();
                log.warn("tcp客户端关闭！");
            }));
        } catch (Exception e) {
            log.error("tcp客户端连接服务端失败", e);
        }
    }


    //public static void sendMessage(String ip, Integer port, String msg) {
    //    byte[] bytes = HexUtil.decodeHex(msg);
    //    try {
    //        if (channelFutureMap.isEmpty()) {
    //            return;
    //        }
    //        tcpChannelFuture.channel()
    //                .writeAndFlush(new DatagramPacket(Unpooled.copiedBuffer(bytes), new InetSocketAddress(ip, port))).sync();
    //        log.debug("发送信号机:{}/{}命令消息:{}", ip, port, msg);
    //    } catch (InterruptedException e) {
    //        log.error("sendMsg is error", e);
    //    }
    //}

    //public static MessageResultPojo sendMessage(String ip, Integer port, String msg, String command, int timeout) {
    //    try {
    //        if (timeout > 0) {
    //            CountDownLatch countDownLatch = new CountDownLatch(1);
    //            String key = StringUtils.join("/", ip, ":", port, "/", command);
    //            long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
    //            MessageResultPojo messageResultPojo = new MessageResultPojo(command, countDownLatch, now, timeout, null);
    //            NettyMessageCache.NETTY_RESULT_TIMEOUT_MAP.put(key, messageResultPojo);
    //            NettyMessageCache.COMMAND_RESULT_SIGN_MAP.put(key, CrossInfoCache.getCrossInfoByIp(ip));
    //            sendMessage(ip, port, msg);
    //            countDownLatch.await(timeout, TimeUnit.MILLISECONDS);
    //            return NettyMessageCache.NETTY_RESULT_TIMEOUT_MAP.remove(key);
    //        }
    //    } catch (InterruptedException e) {
    //        log.error("sendMsg is error", e);
    //    }
    //    return null;
    //}

    @Scheduled(fixedRate = 300)
    public void checkWaitTimeout() {
        long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
        NettyMessageCache.NETTY_RESULT_TIMEOUT_MAP.forEach((k, v) -> {
            commonThreadPoolExecutor.execute(() -> {
                if (now - v.getStartTime() > v.getWaitMillisecond()) {
                    log.warn("wait {}ms [{}] timeout", v.getWaitMillisecond(), k);
                    v.getCountDownLatch().countDown();
                }
            });
        });
    }

    @Scheduled(initialDelay = 1000 * 30, fixedRate = 1000 * 20)
    public void hearBeat() throws Exception {
        try {
            HeartBeatRequest heartBeatRequest = new HeartBeatRequest();
            heartBeatRequest.setMessageType(MessageTypeEnum.HEARTBEAT_STATUS.getType());
            log.error("心跳消息：{}", XMLUtils.convertToXml(heartBeatRequest));
            if (Objects.nonNull(tcpChannelFuture)) {
                tcpChannelFuture.channel().writeAndFlush(XMLUtils.strToHex(XMLUtils.convertToXml(heartBeatRequest)));
            }
        } catch (Exception e) {
            log.error("每20秒心跳发送失败：", e);
            throw new Exception(e);
        }
    }
}
